"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.GetByteSizeVisitor = void 0;
const BaseThrowVisitor_1 = require("../BaseThrowVisitor");
class GetByteSizeVisitor extends BaseThrowVisitor_1.BaseThrowVisitor {
    constructor() {
        super(...arguments);
        this.availableDefinedTypes = new Map();
        this.visitedDefinedTypes = new Map();
        this.definedTypeStack = [];
    }
    registerDefinedTypes(definedTypes) {
        definedTypes.forEach((definedType) => {
            this.availableDefinedTypes.set(definedType.name, definedType);
        });
    }
    visitAccount(account) {
        return account.type.accept(this);
    }
    visitInstruction(instruction) {
        return instruction.args.accept(this);
    }
    visitDefinedType(definedType) {
        if (this.visitedDefinedTypes.has(definedType.name)) {
            return this.visitedDefinedTypes.get(definedType.name);
        }
        this.definedTypeStack.push(definedType.name);
        const child = definedType.type.accept(this);
        this.definedTypeStack.pop();
        this.visitedDefinedTypes.set(definedType.name, child);
        return child;
    }
    visitTypeArray(typeArray) {
        if (typeArray.size.kind !== 'fixed')
            return null;
        const itemSize = typeArray.item.accept(this);
        const arraySize = itemSize !== null ? itemSize * typeArray.size.size : null;
        return typeArray.size.size === 0 ? 0 : arraySize;
    }
    visitTypeDefinedLink(typeDefinedLink) {
        if (typeDefinedLink.dependency !== 'generated') {
            return null;
        }
        const linkedDefinedType = this.availableDefinedTypes.get(typeDefinedLink.name);
        if (!linkedDefinedType) {
            throw new Error(`Cannot find linked defined type ${typeDefinedLink.name}.`);
        }
        // This prevents infinite recursion by using assuming
        // cyclic types don't have a fixed size.
        if (this.definedTypeStack.includes(linkedDefinedType.name)) {
            return null;
        }
        return linkedDefinedType.accept(this);
    }
    visitTypeEnum(typeEnum) {
        if (typeEnum.isScalarEnum())
            return 1;
        const variantSizes = typeEnum.variants.map((v) => v.accept(this));
        const allVariantHaveTheSameFixedSize = variantSizes.every((one, i, all) => one === all[0]);
        return allVariantHaveTheSameFixedSize &&
            variantSizes.length > 0 &&
            variantSizes[0] !== null
            ? variantSizes[0] + 1
            : null;
    }
    visitTypeEnumEmptyVariant() {
        return 0;
    }
    visitTypeEnumStructVariant(typeEnumStructVariant) {
        return typeEnumStructVariant.struct.accept(this);
    }
    visitTypeEnumTupleVariant(typeEnumTupleVariant) {
        return typeEnumTupleVariant.tuple.accept(this);
    }
    visitTypeMap() {
        return null;
    }
    visitTypeOption(typeOption) {
        if (!typeOption.fixed)
            return null;
        const prefixSize = typeOption.prefix.accept(this);
        const itemSize = typeOption.item.accept(this);
        return itemSize !== null ? itemSize + prefixSize : null;
    }
    visitTypeSet() {
        return null;
    }
    visitTypeStruct(typeStruct) {
        return this.sumSizes(typeStruct.fields.map((f) => f.accept(this)));
    }
    visitTypeStructField(typeStructField) {
        return typeStructField.type.accept(this);
    }
    visitTypeTuple(typeTuple) {
        return this.sumSizes(typeTuple.items.map((i) => i.accept(this)));
    }
    visitTypeBool(typeBool) {
        return typeBool.size.accept(this);
    }
    visitTypeBytes() {
        return null;
    }
    visitTypeNumber(typeNumber) {
        return parseInt(typeNumber.format.slice(1), 10) / 8;
    }
    visitTypeNumberWrapper(typeNumberWrapper) {
        return typeNumberWrapper.item.accept(this);
    }
    visitTypePublicKey() {
        return 32;
    }
    visitTypeString(typeString) {
        if (typeString.size.kind !== 'fixed')
            return null;
        return typeString.size.bytes;
    }
    sumSizes(sizes) {
        return sizes.reduce((all, one) => (all === null || one === null ? null : all + one), 0);
    }
}
exports.GetByteSizeVisitor = GetByteSizeVisitor;
//# sourceMappingURL=GetByteSizeVisitor.js.map