"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.GetDefinedTypeHistogramVisitor = void 0;
class GetDefinedTypeHistogramVisitor {
    constructor() {
        this.mode = null;
        this.stackLevel = 0;
    }
    visitRoot(root) {
        return this.mergeHistograms(root.programs.map((program) => program.accept(this)));
    }
    visitProgram(program) {
        return this.mergeHistograms([
            ...program.accounts.map((account) => account.accept(this)),
            ...program.instructions.map((instruction) => instruction.accept(this)),
            ...program.definedTypes.map((type) => type.accept(this)),
        ]);
    }
    visitAccount(account) {
        this.mode = 'account';
        this.stackLevel = 0;
        const histogram = account.type.accept(this);
        this.mode = null;
        return histogram;
    }
    visitInstruction(instruction) {
        this.mode = 'instruction';
        this.stackLevel = 0;
        const histogram = instruction.args.accept(this);
        this.mode = null;
        const subHistograms = instruction.subInstructions.map((ix) => ix.accept(this));
        return this.mergeHistograms([histogram, ...subHistograms]);
    }
    visitDefinedType(definedType) {
        this.mode = 'definedType';
        this.stackLevel = 0;
        const histogram = definedType.type.accept(this);
        this.mode = null;
        return histogram;
    }
    visitError() {
        return {};
    }
    visitTypeArray(typeArray) {
        this.stackLevel += 1;
        const histogram = typeArray.item.accept(this);
        this.stackLevel -= 1;
        return histogram;
    }
    visitTypeDefinedLink(typeDefinedLink) {
        if (typeDefinedLink.dependency !== 'generated') {
            return {};
        }
        return {
            [typeDefinedLink.name]: {
                total: 1,
                inAccounts: Number(this.mode === 'account'),
                inDefinedTypes: Number(this.mode === 'definedType'),
                inInstructionArgs: Number(this.mode === 'instruction'),
                directlyAsInstructionArgs: Number(this.mode === 'instruction' && this.stackLevel <= 2),
            },
        };
    }
    visitTypeEnum(typeEnum) {
        this.stackLevel += 1;
        const histogram = this.mergeHistograms(typeEnum.variants.map((variant) => variant.accept(this)));
        this.stackLevel -= 1;
        return histogram;
    }
    visitTypeEnumEmptyVariant() {
        return {};
    }
    visitTypeEnumStructVariant(typeEnumStructVariant) {
        this.stackLevel += 1;
        const histogram = typeEnumStructVariant.struct.accept(this);
        this.stackLevel -= 1;
        return histogram;
    }
    visitTypeEnumTupleVariant(typeEnumTupleVariant) {
        this.stackLevel += 1;
        const histogram = typeEnumTupleVariant.tuple.accept(this);
        this.stackLevel -= 1;
        return histogram;
    }
    visitTypeMap(typeMap) {
        this.stackLevel += 1;
        const histogram = this.mergeHistograms([
            typeMap.key.accept(this),
            typeMap.value.accept(this),
        ]);
        this.stackLevel -= 1;
        return histogram;
    }
    visitTypeOption(typeOption) {
        this.stackLevel += 1;
        const histogram = typeOption.item.accept(this);
        this.stackLevel -= 1;
        return histogram;
    }
    visitTypeSet(typeSet) {
        this.stackLevel += 1;
        const histogram = typeSet.item.accept(this);
        this.stackLevel -= 1;
        return histogram;
    }
    visitTypeStruct(typeStruct) {
        this.stackLevel += 1;
        const histogram = this.mergeHistograms(typeStruct.fields.map((field) => field.accept(this)));
        this.stackLevel -= 1;
        return histogram;
    }
    visitTypeStructField(typeStructField) {
        this.stackLevel += 1;
        const histogram = typeStructField.type.accept(this);
        this.stackLevel -= 1;
        return histogram;
    }
    visitTypeTuple(typeTuple) {
        this.stackLevel += 1;
        const histogram = this.mergeHistograms(typeTuple.items.map((item) => item.accept(this)));
        this.stackLevel -= 1;
        return histogram;
    }
    visitTypeBool() {
        return {};
    }
    visitTypeBytes() {
        return {};
    }
    visitTypeNumber() {
        return {};
    }
    visitTypeNumberWrapper(typeNumberWrapper) {
        this.stackLevel += 1;
        const histogram = typeNumberWrapper.item.accept(this);
        this.stackLevel -= 1;
        return histogram;
    }
    visitTypePublicKey() {
        return {};
    }
    visitTypeString() {
        return {};
    }
    mergeHistograms(histograms) {
        const result = {};
        histograms.forEach((histogram) => {
            Object.keys(histogram).forEach((key) => {
                if (result[key] === undefined) {
                    result[key] = histogram[key];
                }
                else {
                    result[key].total += histogram[key].total;
                    result[key].inAccounts += histogram[key].inAccounts;
                    result[key].inDefinedTypes += histogram[key].inDefinedTypes;
                    result[key].inInstructionArgs += histogram[key].inInstructionArgs;
                    result[key].directlyAsInstructionArgs +=
                        histogram[key].directlyAsInstructionArgs;
                }
            });
        });
        return result;
    }
}
exports.GetDefinedTypeHistogramVisitor = GetDefinedTypeHistogramVisitor;
//# sourceMappingURL=GetDefinedTypeHistogramVisitor.js.map